@@ -26,7 +26,6 @@ import java.util.concurrent.atomic.AtomicBoolean;
 import javax.servlet.ReadListener;
 
 import org.apache.tomcat.util.buf.B2CConverter;
-import org.apache.tomcat.util.buf.ByteChunk;
 import org.apache.tomcat.util.buf.MessageBytes;
 import org.apache.tomcat.util.buf.UDecoder;
 import org.apache.tomcat.util.http.MimeHeaders;
@@ -36,13 +35,21 @@ import org.apache.tomcat.util.net.ApplicationBufferHandler;
 import org.apache.tomcat.util.res.StringManager;
 
 /**
- * This is a low-level, efficient representation of a server request. Most fields are GC-free, expensive operations are
- * delayed until the user code needs the information. Processing is delegated to modules, using a hook mechanism. This
- * class is not intended for user code - it is used internally by tomcat for processing the request in the most
- * efficient way. Users ( servlets ) can access the information using a facade, which provides the high-level view of
- * the request. Tomcat defines a number of attributes:
+ * This is a low-level, efficient representation of a server request. Most
+ * fields are GC-free, expensive operations are delayed until the  user code
+ * needs the information.
+ *
+ * Processing is delegated to modules, using a hook mechanism.
+ *
+ * This class is not intended for user code - it is used internally by tomcat
+ * for processing the request in the most efficient way. Users ( servlets ) can
+ * access the information using a facade, which provides the high-level view
+ * of the request.
+ *
+ * Tomcat defines a number of attributes:
  * <ul>
- * <li>"org.apache.tomcat.request" - allows access to the low-level request object in trusted applications
+ *   <li>"org.apache.tomcat.request" - allows access to the low-level
+ *       request object in trusted applications
  * </ul>
  *
  * @author James Duncan Davidson [duncan@eng.sun.com]
@@ -93,11 +100,12 @@ public final class Request {
     private final MessageBytes localAddrMB = MessageBytes.newInstance();
 
     private final MimeHeaders headers = new MimeHeaders();
+    private final Map<String,String> trailerFields = new HashMap<>();
 
     /**
      * Path parameters
      */
-    private final Map<String, String> pathParameters = new HashMap<>();
+    private final Map<String,String> pathParameters = new HashMap<>();
 
     /**
      * Notes.
@@ -138,18 +146,17 @@ public final class Request {
     private final MessageBytes remoteUser = MessageBytes.newInstance();
     private boolean remoteUserNeedsAuthorization = false;
     private final MessageBytes authType = MessageBytes.newInstance();
-    private final HashMap<String, Object> attributes = new HashMap<>();
+    private final HashMap<String,Object> attributes = new HashMap<>();
 
     private Response response;
     private volatile ActionHook hook;
 
-    private long bytesRead = 0;
+    private long bytesRead=0;
     // Time of the request - useful to avoid repeated calls to System.currentTime
     private long startTime = -1;
-    private long threadId = 0;
     private int available = 0;
 
-    private final RequestInfo reqProcessorMX = new RequestInfo(this);
+    private final RequestInfo reqProcessorMX=new RequestInfo(this);
 
     private boolean sendfile = true;
 
@@ -159,8 +166,9 @@ public final class Request {
     private Exception errorException = null;
 
     /*
-     * State for non-blocking output is maintained here as it is the one point easily reachable from the
-     * CoyoteInputStream and the CoyoteAdapter which both need access to state.
+     * State for non-blocking output is maintained here as it is the one point
+     * easily reachable from the CoyoteInputStream and the CoyoteAdapter which
+     * both need access to state.
      */
     volatile ReadListener listener;
     // Ensures listener is only fired after a call is isReady()
@@ -176,17 +184,20 @@ public final class Request {
 
     public void setReadListener(ReadListener listener) {
         if (listener == null) {
-            throw new NullPointerException(sm.getString("request.nullReadListener"));
+            throw new NullPointerException(
+                    sm.getString("request.nullReadListener"));
         }
         if (getReadListener() != null) {
-            throw new IllegalStateException(sm.getString("request.readListenerSet"));
+            throw new IllegalStateException(
+                    sm.getString("request.readListenerSet"));
         }
         // Note: This class is not used for HTTP upgrade so only need to test
-        // for async
+        //       for async
         AtomicBoolean result = new AtomicBoolean(false);
         action(ActionCode.ASYNC_IS_ASYNC, result);
         if (!result.get()) {
-            throw new IllegalStateException(sm.getString("request.notAsync"));
+            throw new IllegalStateException(
+                    sm.getString("request.notAsync"));
         }
 
         this.listener = listener;
@@ -209,7 +220,7 @@ public final class Request {
                 fireListener = true;
             }
             action(ActionCode.DISPATCH_READ, null);
-            if (!isRequestThread()) {
+            if (!ContainerThreadMarker.isContainerThread()) {
                 // Not on a container thread so need to execute the dispatch
                 action(ActionCode.DISPATCH_EXECUTE, null);
             }
@@ -270,6 +281,18 @@ public final class Request {
     }
 
 
+    public boolean isTrailerFieldsReady() {
+        AtomicBoolean result = new AtomicBoolean(false);
+        action(ActionCode.IS_TRAILER_FIELDS_READY, result);
+        return result.get();
+    }
+
+
+    public Map<String,String> getTrailerFields() {
+        return trailerFields;
+    }
+
+
     public UDecoder getURLDecoder() {
         return urlDecoder;
     }
@@ -302,9 +325,11 @@ public final class Request {
     }
 
     /**
-     * Get the "virtual host", derived from the Host: header associated with this request.
+     * Get the "virtual host", derived from the Host: header associated with
+     * this request.
      *
-     * @return The buffer holding the server name, if any. Use isNull() to check if there is no value set.
+     * @return The buffer holding the server name, if any. Use isNull() to check
+     *         if there is no value set.
      */
     public MessageBytes serverName() {
         return serverNameMB;
@@ -314,8 +339,8 @@ public final class Request {
         return serverPort;
     }
 
-    public void setServerPort(int serverPort) {
-        this.serverPort = serverPort;
+    public void setServerPort(int serverPort ) {
+        this.serverPort=serverPort;
     }
 
     public MessageBytes remoteAddr() {
@@ -338,19 +363,19 @@ public final class Request {
         return localAddrMB;
     }
 
-    public int getRemotePort() {
+    public int getRemotePort(){
         return remotePort;
     }
 
-    public void setRemotePort(int port) {
+    public void setRemotePort(int port){
         this.remotePort = port;
     }
 
-    public int getLocalPort() {
+    public int getLocalPort(){
         return localPort;
     }
 
-    public void setLocalPort(int port) {
+    public void setLocalPort(int port){
         this.localPort = port;
     }
 
@@ -360,8 +385,9 @@ public final class Request {
     /**
      * Get the character encoding used for this request.
      *
-     * @return The value set via {@link #setCharacterEncoding(String)} or if no call has been made to that method try to
-     *             obtain if from the content type.
+     * @return The value set via {@link #setCharset(Charset)} or if no
+     *         call has been made to that method try to obtain if from the
+     *         content type.
      */
     public String getCharacterEncoding() {
         if (characterEncoding == null) {
@@ -375,10 +401,12 @@ public final class Request {
     /**
      * Get the character encoding used for this request.
      *
-     * @return The value set via {@link #setCharacterEncoding(String)} or if no call has been made to that method try to
-     *             obtain if from the content type.
+     * @return The value set via {@link #setCharset(Charset)} or if no
+     *         call has been made to that method try to obtain if from the
+     *         content type.
      *
-     * @throws UnsupportedEncodingException If the user agent has specified an invalid character encoding
+     * @throws UnsupportedEncodingException If the user agent has specified an
+     *         invalid character encoding
      */
     public Charset getCharset() throws UnsupportedEncodingException {
         if (charset == null) {
@@ -386,25 +414,12 @@ public final class Request {
             if (characterEncoding != null) {
                 charset = B2CConverter.getCharset(characterEncoding);
             }
-        }
+         }
 
         return charset;
     }
 
 
-    /**
-     * @param enc The new encoding
-     *
-     * @throws UnsupportedEncodingException If the encoding is invalid
-     *
-     * @deprecated This method will be removed in Tomcat 9.0.x
-     */
-    @Deprecated
-    public void setCharacterEncoding(String enc) throws UnsupportedEncodingException {
-        setCharset(B2CConverter.getCharset(enc));
-    }
-
-
     public void setCharset(Charset charset) {
         this.charset = charset;
         this.characterEncoding = charset.name();
@@ -426,7 +441,7 @@ public final class Request {
     }
 
     public long getContentLengthLong() {
-        if (contentLength > -1) {
+        if( contentLength > -1 ) {
             return contentLength;
         }
 
@@ -459,7 +474,7 @@ public final class Request {
 
 
     public void setContentType(MessageBytes mb) {
-        contentTypeMB = mb;
+        contentTypeMB=mb;
     }
 
 
@@ -530,15 +545,15 @@ public final class Request {
     // -------------------- Other attributes --------------------
     // We can use notes for most - need to discuss what is of general interest
 
-    public void setAttribute(String name, Object o) {
-        attributes.put(name, o);
+    public void setAttribute( String name, Object o ) {
+        attributes.put( name, o );
     }
 
-    public HashMap<String, Object> getAttributes() {
+    public HashMap<String,Object> getAttributes() {
         return attributes;
     }
 
-    public Object getAttribute(String name) {
+    public Object getAttribute(String name ) {
         return attributes.get(name);
     }
 
@@ -601,34 +616,14 @@ public final class Request {
 
 
     /**
-     * Read data from the input buffer and put it into a byte chunk. The buffer is owned by the protocol implementation
-     * - it will be reused on the next read. The Adapter must either process the data in place or copy it to a separate
-     * buffer if it needs to hold it. In most cases this is done during byte-&gt;char conversions or via InputStream.
-     * Unlike InputStream, this interface allows the app to process data in place, without copy.
+     * Read data from the input buffer and put it into ApplicationBufferHandler.
      *
-     * @param chunk The destination to which to copy the data
-     *
-     * @return The number of bytes copied
-     *
-     * @throws IOException If an I/O error occurs during the copy
-     *
-     * @deprecated Unused. Will be removed in Tomcat 9. Use {@link #doRead(ApplicationBufferHandler)}
-     */
-    @Deprecated
-    public int doRead(ByteChunk chunk) throws IOException {
-        int n = inputBuffer.doRead(chunk);
-        if (n > 0) {
-            bytesRead += n;
-        }
-        return n;
-    }
-
-
-    /**
-     * Read data from the input buffer and put it into ApplicationBufferHandler. The buffer is owned by the protocol
-     * implementation - it will be reused on the next read. The Adapter must either process the data in place or copy it
-     * to a separate buffer if it needs to hold it. In most cases this is done during byte-&gt;char conversions or via
-     * InputStream. Unlike InputStream, this interface allows the app to process data in place, without copy.
+     * The buffer is owned by the protocol implementation - it will be reused on
+     * the next read. The Adapter must either process the data in place or copy
+     * it to a separate buffer if it needs to hold it. In most cases this is
+     * done during byte-&gt;char conversions or via InputStream. Unlike
+     * InputStream, this interface allows the app to process data in place,
+     * without copy.
      *
      * @param handler The destination to which to copy the data
      *
@@ -643,7 +638,7 @@ public final class Request {
 
         int n = inputBuffer.doRead(handler);
         if (n > 0) {
-            bytesRead += n;
+            bytesRead+=n;
         }
         return n;
     }
@@ -652,7 +647,8 @@ public final class Request {
     // -------------------- Error tracking --------------------
 
     /**
-     * Set the error Exception that occurred during the writing of the response processing.
+     * Set the error Exception that occurred during the writing of the response
+     * processing.
      *
      * @param ex The exception that occurred
      */
@@ -691,41 +687,33 @@ public final class Request {
         this.startTime = startTime;
     }
 
-    public long getThreadId() {
-        return threadId;
-    }
-
-    public void clearRequestThread() {
-        threadId = 0;
-    }
-
-    public void setRequestThread() {
-        threadId = Thread.currentThread().getId();
-    }
-
-    public boolean isRequestThread() {
-        return Thread.currentThread().getId() == threadId;
-    }
-
     // -------------------- Per-Request "notes" --------------------
 
 
     /**
-     * Used to store private data. Thread data could be used instead - but if you have the req, getting/setting a note
-     * is just a array access, may be faster than ThreadLocal for very frequent operations. Example use: Catalina
-     * CoyoteAdapter: ADAPTER_NOTES = 1 - stores the HttpServletRequest object ( req/res) To avoid conflicts, note in
-     * the range 0 - 8 are reserved for the servlet container ( catalina connector, etc ), and values in 9 - 16 for
-     * connector use. 17-31 range is not allocated or used.
+     * Used to store private data. Thread data could be used instead - but
+     * if you have the req, getting/setting a note is just an array access, may
+     * be faster than ThreadLocal for very frequent operations.
      *
-     * @param pos   Index to use to store the note
+     *  Example use:
+     *   Catalina CoyoteAdapter:
+     *      ADAPTER_NOTES = 1 - stores the HttpServletRequest object ( req/res)
+     *
+     *   To avoid conflicts, note in the range 0 - 8 are reserved for the
+     *   servlet container ( catalina connector, etc ), and values in 9 - 16
+     *   for connector use.
+     *
+     *   17-31 range is not allocated or used.
+     *
+     * @param pos Index to use to store the note
      * @param value The value to store at that index
      */
-    public void setNote(int pos, Object value) {
+    public final void setNote(int pos, Object value) {
         notes[pos] = value;
     }
 
 
-    public Object getNote(int pos) {
+    public final Object getNote(int pos) {
         return notes[pos];
     }
 
@@ -734,7 +722,7 @@ public final class Request {
 
 
     public void recycle() {
-        bytesRead = 0;
+        bytesRead=0;
 
         contentLength = -1;
         contentTypeMB = null;
@@ -742,8 +730,9 @@ public final class Request {
         characterEncoding = null;
         expectation = false;
         headers.recycle();
+        trailerFields.clear();
         serverNameMB.recycle();
-        serverPort = -1;
+        serverPort=-1;
         localAddrMB.recycle();
         localNameMB.recycle();
         localPort = -1;
@@ -781,10 +770,9 @@ public final class Request {
         allDataReadEventSent.set(false);
 
         startTime = -1;
-        threadId = 0;
     }
 
-    // -------------------- Info --------------------
+    // -------------------- Info  --------------------
     public void updateCounters() {
         reqProcessorMX.updateCounters();
     }
@@ -798,12 +786,13 @@ public final class Request {
     }
 
     public boolean isProcessing() {
-        return reqProcessorMX.getStage() == org.apache.coyote.Constants.STAGE_SERVICE;
+        return reqProcessorMX.getStage()==org.apache.coyote.Constants.STAGE_SERVICE;
     }
 
     /**
-     * Parse the character encoding from the specified content type header. If the content type is null, or there is no
-     * explicit character encoding, <code>null</code> is returned.
+     * Parse the character encoding from the specified content type header.
+     * If the content type is null, or there is no explicit character encoding,
+     * <code>null</code> is returned.
      *
      * @param contentType a content type header
      */
@@ -822,7 +811,8 @@ public final class Request {
             encoding = encoding.substring(0, end);
         }
         encoding = encoding.trim();
-        if ((encoding.length() > 2) && (encoding.startsWith("\"")) && (encoding.endsWith("\""))) {
+        if ((encoding.length() > 2) && (encoding.startsWith("\""))
+            && (encoding.endsWith("\""))) {
             encoding = encoding.substring(1, encoding.length() - 1);
         }
 
